/*
Name :  compute.c
Zweck:  Wahrscheinlichkeitsberechnung in einem Verteilten System
Autor:  Andreas Seifert
Datum:
Update:
*/
/*-------------------------------------------------------------------------*/


/****************************************************************************/

#include "main.h"
#include "graph.h"
#include "compute.h"

/****************************************************************************/



/*
 * AU: Wahrscheinlichkeitsberechnung
 * ST: computeReli()               Verfuegbarkeit eines verteilten Systems 
 *                                 berechnen
 *     findpath()                  Bestehen eines Weges zwischen Rechnern 
 *                                 pruefen
 *     successfullVQM()            Pruefung auf eine Verbindung zwischen den
 *                                 Mitgliedern einer Virtuellen Quorenmenge
 *     connectionExists()          Pruefung auf Verbindung zwischen einer 
 *                                 Menge von Rechnern
 *     addProbability()            Wahrscheinlichkeit einer Konstellation 
 *                                 berechnen
 *     markGraph()                 Bestimmte Konstellation eines Verteiltes 
 *                                 Systems herstellen
 *     permutArray()               Konstellation des verteilten Systems 
 *                                 permutieren
 *     computeCorrespondingArray() Elemente eines verteilten Systems in einem
 *                                 Vektor abbilden
 *     correspondingArc()          Bestimme die entgegengesetzte Verbindung
 *     initializeNet()             Verteiltes System initialisieren  
 */

/**************************************************************************/
/*                     Beginn der Funktionen                              */
/**************************************************************************/


void computeReli(ObjectList, service)
     NODE  *ObjectList; /* Liste der Rechner des Netzes */
     int service;       /* Operation/Dienst */
/*
 * AU: Verfuegbarkeit eines verteilten Systems berechnen durch
 *     Simulierung aller 2^(#Rechner + #Verbindungen) moeglichen
 *     Systemzustaende
 * SU: initializeNet()
 *     computeCorrespondingArray()
 *     markGraph()
 *     successfullVQM()
 *     addProbability()
 *     permutArray() 
 */
{
 int    size;         /* Anzahl der Elemente des verteilten Systems */
 int    i, remainder; /* Laufvariable und Divisionsrest zur Anzeige des
                       * Bearbeitungsstandes */
 double endloop;      /* Schleifenende = 2^size */

 /* Netz initialisieren */
 initializeNet(ObjectList);

 /* Repraesentation des Netzes als Vektor erstellen */
 computeCorrespondingArray(ObjectList, &size);

 /* Anzahl moeglicher Systemzustaende ermitteln */
 /* exp2() unbekannt (!?!): endloop = exp2((double)size); */
 endloop = pow((double)2.0,(double)size);

 for (i = 0; i < endloop; i++) { /* Zustaende der Netztopologie erzeugen */
      markGraph(&size); /* Zustand des Vektors auf Netz abbilden */

      remainder = i % 10000; /* Divisionsrest == 0 veranlasst die Anzeige des
                               * Bearbeitungsstandes */

      if (remainder == 0) ActualPercent(endloop, i, service);

      /* Koennen die Mitglieder einer virtuellen Gruppe im verteilten
       * System zusammenarbeiten, ist die W. zu ermitteln und der 
       * Geamtverfuegbarkeit des verteilten Systems hinzuzufuegen */ 
      if (successfullVQM(service, ObjectList)) addProbability(&size, service);
      
      /* Naechste Konstellation erzeugen */
      permutArray(&size);
     }

 /* Ergebnis der Berechnung in globaler Variablen sichern */
 data.reliability[service] = data.reliability[service] / endloop;

 XtFree((char *)list);
 list = NULL; /* wird evt. spaeter noch mal als Argument zu XtRealloc() benutzt */
}


void findpath(ptrFromNodeObject)
     NODE *ptrFromNodeObject;
/*
 * AU: Alle von einem Rechner aus erreichbaren Rechner markieren
 * SU: findpath();
 */
{
 ARC *arcToNeighbour;

 if ((ptrFromNodeObject->nodeStatus == non_visited) &&   /* Knoten noch */
     (ptrFromNodeObject->nodeCondition == working)) {    /* nicht besucht */
                                                         /* Knoten aktiv? */
     ptrFromNodeObject->nodeStatus = visited;            /* markieren */

     for (arcToNeighbour = ptrFromNodeObject->arcptr;    /* fuer jeden */
         arcToNeighbour != NULL;                         /* Nachbarknoten */
         arcToNeighbour = arcToNeighbour->nextarc) {

         if ((arcToNeighbour->arcCondition == working) &&
             (arcToNeighbour->arcType == other_object)) {
             arcToNeighbour->arcType = visited_object;
             findpath(arcToNeighbour->ndptr);   
             }
         }
     }
}


Bool successfullVQM(service, ObjectList)
     int  service;
     NODE *ObjectList;
/*
 * AU: Pruefung auf eine Verbindung zwischen den Mitgliedern 
 *     einer Virtuellen Quorenmenge
 */
{
 VQM_SET *ptrSet;

 if (ptrVQMSet[service] == NULL) return False; /* leere VQM */

 /* Pruefe Verbindung fuer jede Quorenmenge */
 for (ptrSet = ptrVQMSet[service]; ptrSet != NULL;
      ptrSet = ptrSet->nextVQMSetHead) {
     if (connectionExists(ptrSet, ObjectList)) return True;
     } 
 /* es existiert keine Verbindung */
 return False;
}



Bool connectionExists(ptrSet, ObjectList)
     VQM_SET *ptrSet;
     NODE    *ObjectList;
/*
 * AU: Pruefung auf Verbindung zwischen einer Menge von Rechnern
 */
{
 VQM_SET_ELEMENT *ptrElement, *ptrNextElement; /* Rechner zwischen denen
                                                * Verbindung gesucht wird */
 NODE            *ptrFromNodeObject, *ptrToNodeObject; /* entsprechende Rechner
                                                        * der Netztopologie */
 

 ptrElement = ptrSet->firstVQMSetElement; /* erstes Element der Gruppe */

 /* Fall einer leeren Gruppe */
 if (ptrElement == NULL) return False;
 else ptrFromNodeObject = /* bestimme Rechner in der Netztopologie */
         (NODE *)correspondingNetNode(ptrElement->identVQMSet);


 ptrNextElement = ptrElement->nextVQMSetElement; /* Rechner zu dem eine Verbindung
                                                  * gesucht wird */

 /* Fall einer einelementigen Gruppe:
  * in diesem Fall reicht es aus, die Funktionsfaehigkeit des einzigen
  * Elementes zu pruefen.
  */
 if (ptrNextElement == NULL) { /* einelementige Gruppe */
    if (ptrFromNodeObject->nodeCondition == failed) return False;
    else                                            return True;
    }


 /* Fall einer mehrelementigen Gruppe */
 ptrToNodeObject = /* bestimme Rechner in der Netztopologie */
    (NODE *)correspondingNetNode(ptrNextElement->identVQMSet);

 do {  /* Pruefung einer neuen Gruppe */ 

    do {  /* Pruefung innerhalb einer Gruppe mit dem naechsten Startelement */
       initializeNet(ObjectList); /* markiere alle Rechner als nicht besucht */

       findpath(ptrFromNodeObject); /* markiere alle erreichbaren Knoten */

       /* pruefe ob der Zielrechner markiert werden konnte */
       if (ptrToNodeObject->nodeStatus == non_visited) return False;

       /* Erreichbarkeit zum naechsten Element der Gruppe pruefen */
       ptrNextElement = ptrNextElement->nextVQMSetElement;

       if (ptrNextElement != NULL) {
          ptrToNodeObject = /* bestimme Rechner in der Netztopologie */
             (NODE *)correspondingNetNode(ptrNextElement->identVQMSet);
          }
       }
    while (ptrNextElement != NULL);

    /* Zeiger auf das naechste Element der Gruppe, von dem aus die Ereichbarkeit
     * zu den anderen Elementen der Gruppe geprueft werden soll */
    ptrElement = ptrElement->nextVQMSetElement;

    if (ptrElement->nextVQMSetElement == NULL) ptrElement = NULL; 
    else {
         ptrNextElement = ptrElement->nextVQMSetElement;

         ptrToNodeObject = /* bestimme Rechner in der Netztopologie */
            (NODE *)correspondingNetNode(ptrNextElement->identVQMSet);

         ptrFromNodeObject = /* bestimme Rechner in der Netztopologie */
            (NODE *)correspondingNetNode(ptrElement->identVQMSet);
         }
    }
 while (ptrElement != NULL); 

 return True; 
}




void addProbability(size, service)
     int *size;   /* Groesse des Vektors */
     int service; /* Dienst */
/*
 * AU: Wahrscheinlichkeit einer Konstellation berechnen und im Erfolgsfall
 *     der Gesamtwhrscheinlichkeit hinzufuegen
 * NB: Multiplikation der Einzelwahrscheinlichkeiten mit 2 wg. Rundung
 */
{
 int    i;     /* Laufvariable ueber alle Komponenten des verteilten Systems */
 double value; /* Wahrscheinlichkeit fuer den gegenwaertig untersuchten Zustand */

 value = 1; 

 for (i = 0; i < *size; i++) {
      switch (list[i].condition)
      {
        case failed : /* Element des verteilten Systems ausgefallen */
             value = value * list[i].reliability * 2;
             break;

        case working : /* Element des verteilten Systems einsatzbereit */
             value = value * (1 - list[i].reliability) * 2;
             break;
	default: /* "warning: enumeration value `endstate' not handled in switch" */
             fprintf(stderr, "%s:%i:%s: Unzulaessiger/nicht abgefangener Zustand: %i\n",
	             __FILE__, __LINE__, __FUNCTION__, list[i].condition);
	     abort();
      }
     }
 data.reliability[service] = data.reliability[service] + value;
}



void markGraph(size)
     int *size; /* Groesse des Vektors */
/*
 * AU: Bestimmte Konstellation eines Verteiltes Systems herstellen
 *     und die durch den Vektor dargestellten Zustaende jeder einzelnen
 *     Komponente auf das Netz uebertragen
 */
{
 int  i;

 for (i = 0; i < *size; i++) { /* Eintrag symbolisiert eine Verbindung */
      if (list[i].pointtonode == NULL) { 
          list[i].pointtoarcs[0]->arcCondition = list[i].condition;
          list[i].pointtoarcs[1]->arcCondition = list[i].condition;
         }
      else list[i].pointtonode->nodeCondition = list[i].condition;
     }
}


void permutArray(size)
     int *size;
/*
 * AU: Konstellation des verteilten Systems permutieren, wobei
 *     der Vektor wie ein binaerer Zaehler benutzt wird
 */
{
 int k;
 int l = 0;  

 while ((list[l].condition == failed) && (l < *size)) l++;

 if (list[l].condition == working) list[l].condition = failed;

 for (k = l - 1; k >= 0; k--) list[k].condition = working;
}



void computeCorrespondingArray(ObjectList, size)
     NODE *ObjectList;
     int  *size;
/*
 * AU: Elemente eines verteilten Systems in einem Vektor abbilden
 * SU: demand()
 */
{
 ARC   *arc;       /* Verbindung des verteilten Systems */
 NODE  *node;      /* Rechner des verteilten Systems */
 int   offset = 1;

 /* Durchlaufe die Liste der Rechner des Systems */
 for (node = ObjectList; node != NULL; node = node->nextnode) {

      /* neues Element anfuegen und die alten Eintraege kopieren */
      list = (VECTOR *)XtRealloc((char *)list, offset * sizeof(VECTOR)); 

      demand(list, allocation in computeCorrespondingArray failed);

      list[offset-1].pointtonode    = node;
      list[offset-1].pointtoarcs[0] = NULL;
      list[offset-1].pointtoarcs[1] = NULL;
      list[offset-1].condition      = working;
      list[offset-1].reliability    = node->reliability;

      offset++; /* Anzahl der Elemente um einen Rechner anpassen */

      /* Die Verbindungen des Rechners als Elemente des Vektors aufnehmen */
      for (arc = node->arcptr; arc != NULL; arc = arc->nextarc) { 
           if (arc->arcType == other_object) { /* nur die Verbindungen sind
                                                * aufzunehmen, die nicht bereits
                                                * als Verbindung eines anderen
                                                * Rechners enthalten sind */

               /* neues Element anfuegen und die alten Eintraege kopieren */
               list = (VECTOR *)XtRealloc((char *)list, offset * sizeof(VECTOR));

               demand(list, allocation in computeCorrespondingArray failed);

               list[offset-1].pointtonode    = NULL;
               list[offset-1].pointtoarcs[0] = arc;
               list[offset-1].pointtoarcs[1] = (ARC *)correspondingArc(arc);
               list[offset-1].condition      = working;
               list[offset-1].reliability    = arc->reliability;
               list[offset-1].pointtoarcs[0]->arcType = visited_object;
               list[offset-1].pointtoarcs[1]->arcType = visited_object;

               offset++; /* Anzahl der Elemente um einen Rechner anpassen */
              }
          }
     }
 *size = offset - 1;
}



ARC *correspondingArc(arc)
    ARC *arc;
/*
 * AU: Bestimme die entgegengesetzte Verbindung
 * NB: Rueckgabewert NULL bei erfolgloser Suche, sonst Zeiger auf Verbindung
 */
{
 NODE *node = arc->ndptr;
 ARC  *helparc;

 for (helparc = node->arcptr; helparc != NULL; helparc = helparc->nextarc) {
      if (strncmp(arc->fatherndptr->ident, helparc->ndptr->ident, 10) == 0) 
          return helparc;
     }
 return NULL;
}


void initializeNet(ObjectList)
     NODE *ObjectList; /* Liste der Rechner des verteilten Systems */
/*
 * AU: Verteiltes System initialisieren  und Rechner als nicht besucht
 *     markieren und die Verbindungen als noch nicht benutzt
 */
{
 ARC   *arc;
 NODE  *node;

 for (node = ObjectList; node != NULL; node = node->nextnode) {
      node->nodeStatus = non_visited; /* Knoten nicht besucht */
      for (arc = node->arcptr; arc != NULL; arc = arc->nextarc)
           arc->arcType = other_object; /* Verbindung nicht benutzt */
     }
}


/***************************************************************************/
/*                    Ende der Funktionen                                  */
/***************************************************************************/
